"""
Callable Symbolic Expressions

EXAMPLES:

When you do arithmetic with::

    sage: f(x, y, z) = sin(x+y+z)
    sage: g(x, y) = y + 2*x
    sage: f + g
    (x, y, z) |--> 2*x + y + sin(x + y + z)

::

    sage: f(x, y, z) = sin(x+y+z)
    sage: g(w, t) = cos(w - t)
    sage: f + g
    (t, w, x, y, z) |--> cos(-t + w) + sin(x + y + z)

::

    sage: f(x, y, t) = y*(x^2-t)
    sage: g(x, y, w) = x + y - cos(w)
    sage: f*g
    (x, y, t, w) |--> (x^2 - t)*(x + y - cos(w))*y

::

    sage: f(x,y, t) = x+y
    sage: g(x, y, w) = w + t
    sage: f + g
    (x, y, t, w) |--> t + w + x + y

TESTS:

The arguments in the definition must be symbolic variables (:issue:`10747`)::

    sage: f(1)=2
    Traceback (most recent call last):
    ...
    SyntaxError: can...t assign to function call...

    sage: f(x,1)=2
    Traceback (most recent call last):
    ...
    SyntaxError: can...t assign to function call...

    sage: f(1,2)=3
    Traceback (most recent call last):
    ...
    SyntaxError: can...t assign to function call...

    sage: f(1,2)=x
    Traceback (most recent call last):
    ...
    SyntaxError: can...t assign to function call...

    sage: f(x,2)=x
    Traceback (most recent call last):
    ...
    SyntaxError: can...t assign to function call...
"""
from sage.misc.lazy_import import lazy_import
from sage.symbolic.ring import SymbolicRing, SR
from sage.categories.pushout import ConstructionFunctor
from sage.structure.factory import UniqueFactory


######################################################################
#  Callable functions
######################################################################

class CallableSymbolicExpressionFunctor(ConstructionFunctor):
    def __init__(self, arguments):
        """
        A functor which produces a CallableSymbolicExpressionRing from
        the SymbolicRing.

        EXAMPLES::

            sage: from sage.symbolic.callable import CallableSymbolicExpressionFunctor
            sage: x,y = var('x,y')
            sage: f = CallableSymbolicExpressionFunctor((x,y)); f
            CallableSymbolicExpressionFunctor(x, y)
            sage: f(SR)
            Callable function ring with arguments (x, y)

            sage: loads(dumps(f))
            CallableSymbolicExpressionFunctor(x, y)
        """
        self._arguments = arguments
        from sage.categories.rings import Rings
        self.rank = 3
        ConstructionFunctor.__init__(self, Rings(), Rings())

    def __repr__(self):
        """
        EXAMPLES::

            sage: from sage.symbolic.callable import CallableSymbolicExpressionFunctor
            sage: x,y = var('x,y')
            sage: CallableSymbolicExpressionFunctor((x,y))
            CallableSymbolicExpressionFunctor(x, y)
        """
        return "CallableSymbolicExpressionFunctor%s" % repr(self.arguments())

    def merge(self, other):
        """
        EXAMPLES::

            sage: from sage.symbolic.callable import CallableSymbolicExpressionFunctor
            sage: x,y = var('x,y')
            sage: a = CallableSymbolicExpressionFunctor((x,))
            sage: b = CallableSymbolicExpressionFunctor((y,))
            sage: a.merge(b)
            CallableSymbolicExpressionFunctor(x, y)
        """
        arguments = self.unify_arguments(other)
        return CallableSymbolicExpressionFunctor(arguments)

    def __call__(self, R):
        """
        EXAMPLES::

            sage: from sage.symbolic.callable import CallableSymbolicExpressionFunctor
            sage: x,y = var('x,y')
            sage: a = CallableSymbolicExpressionFunctor((x,y))
            sage: a(SR)
            Callable function ring with arguments (x, y)
        """
        if R is not SR:
            raise ValueError("can only make callable symbolic expression rings from the Symbolic Ring")
        return CallableSymbolicExpressionRing(self.arguments())

    def arguments(self):
        """
        EXAMPLES::

            sage: from sage.symbolic.callable import CallableSymbolicExpressionFunctor
            sage: x,y = var('x,y')
            sage: a = CallableSymbolicExpressionFunctor((x,y))
            sage: a.arguments()
            (x, y)
        """
        return self._arguments

    def unify_arguments(self, x):
        r"""
        Take the variable list from another ``CallableSymbolicExpression``
        object and compare it with the current ``CallableSymbolicExpression``
        object's variable list, combining them according to the following rules:

        Let ``a`` be ``self``'s variable list, let ``b`` be ``y``'s
        variable list.

        #. If ``a == b``, then the variable lists are
           identical, so return that variable list.

        #. If ``a`` `\neq` ``b``, then check if the first `n` items in
           ``a`` are the first `n` items in ``b``, or vice versa. If
           so, return a list with these `n` items, followed by the
           remaining items in ``a`` and ``b`` sorted together in
           alphabetical order.


        .. NOTE::

           When used for arithmetic between
           ``CallableSymbolicExpression``'s, these rules ensure that
           the set of ``CallableSymbolicExpression``'s will have
           certain properties. In particular, it ensures that the set
           is a *commutative* ring, i.e., the order of the input
           variables is the same no matter in which order arithmetic
           is done.

        INPUT:

        - ``x`` -- a ``CallableSymbolicExpression``

        OUTPUT: a tuple of variables

        EXAMPLES::

            sage: from sage.symbolic.callable import CallableSymbolicExpressionFunctor
            sage: x,y = var('x,y')
            sage: a = CallableSymbolicExpressionFunctor((x,))
            sage: b = CallableSymbolicExpressionFunctor((y,))
            sage: a.unify_arguments(b)
            (x, y)

        AUTHORS:

        - Bobby Moretti: thanks to William Stein for the rules
        """
        a = self.arguments()
        b = x.arguments()

        # Rule #1
        if [str(y) for y in a] == [str(z) for z in b]:
            return a

        # Rule #2
        new_list = []
        done = False
        i = 0
        while not done and i < min(len(a), len(b)):
            if repr(a[i]) == repr(b[i]):
                new_list.append(a[i])
                i += 1
            else:
                done = True

        temp = set()
        # Sorting remaining variables.
        for j in range(i, len(a)):
            if a[j] not in temp:
                temp.add(a[j])

        for j in range(i, len(b)):
            if b[j] not in temp:
                temp.add(b[j])

        new_list.extend(sorted(temp, key=repr))
        return tuple(new_list)


class CallableSymbolicExpressionRing_class(SymbolicRing):
    def __init__(self, arguments):
        """
        EXAMPLES:

        We verify that coercion works in the case where ``x`` is not an
        instance of SymbolicExpression, but its parent is still the
        SymbolicRing::

            sage: f(x) = 1
            sage: f*e
            x |--> e

        TESTS::

            sage: TestSuite(f.parent()).run(skip=['_test_divides'])
        """
        self._arguments = arguments
        SymbolicRing.__init__(self, SR)
        self._populate_coercion_lists_(coerce_list=[SR])
        self.symbols = SR.symbols  # Use the same list of symbols as SR

    def _coerce_map_from_(self, R):
        """
        EXAMPLES::

            sage: f(x,y) = x^2 + y
            sage: g(x,y,z) = x + y + z
            sage: f.parent().has_coerce_map_from(g.parent())
            False
            sage: g.parent().has_coerce_map_from(f.parent())
            True
        """
        if isinstance(R, CallableSymbolicExpressionRing_class):
            args = self.arguments()
            return all(a in args for a in R.arguments())
        return SymbolicRing._coerce_map_from_(self, R)

    def construction(self):
        """
        EXAMPLES::

            sage: f(x,y) = x^2 + y
            sage: f.parent().construction()
            (CallableSymbolicExpressionFunctor(x, y), Symbolic Ring)
        """
        return (CallableSymbolicExpressionFunctor(self.arguments()), SR)

    def _element_constructor_(self, x):
        """
        TESTS::

            sage: f(x) = x+1; g(y) = y+1
            sage: f.parent()(g)
            x |--> y + 1
            sage: g.parent()(f)
            y |--> x + 1
            sage: f(x) = x+2*y; g(y) = y+3*x
            sage: f.parent()(g)
            x |--> 3*x + y
            sage: g.parent()(f)
            y |--> x + 2*y
        """
        return SymbolicRing._element_constructor_(self, x)

    def _repr_(self):
        """
        String representation of ring of callable symbolic expressions.

        EXAMPLES::

            sage: R = CallableSymbolicExpressionRing(var('x,y,theta'))
            sage: R._repr_()
            'Callable function ring with arguments (x, y, theta)'

        We verify that :issue:`12298` has been fixed::

            sage: S = CallableSymbolicExpressionRing([var('z')])
            sage: S._repr_()
            'Callable function ring with argument z'
        """
        if len(self._arguments) == 0:
            return "Callable function ring with no named arguments"
        if len(self._arguments) == 1:
            return "Callable function ring with argument {}".format(self._arguments[0])
        return "Callable function ring with arguments {}".format(self._arguments)

    def arguments(self):
        """
        Return the arguments of ``self``.

        The order that the variables appear in ``self.arguments()`` is
        the order that is used in evaluating the elements of ``self``.

        EXAMPLES::

            sage: x,y = var('x,y')
            sage: f(x,y) = 2*x+y
            sage: f.parent().arguments()
            (x, y)
            sage: f(y,x) = 2*x+y
            sage: f.parent().arguments()
            (y, x)
        """
        return self._arguments

    args = arguments

    def _repr_element_(self, x):
        """
        Return the string representation of the Expression ``x``.

        EXAMPLES::

            sage: f(y,x) = x + y
            sage: f
            (y, x) |--> x + y
            sage: f.parent()
            Callable function ring with arguments (y, x)
        """
        args = self.arguments()
        repr_x = SymbolicRing._repr_element_(self, x)
        if len(args) == 1:
            return "%s |--> %s" % (args[0], repr_x)
        args = ", ".join(map(str, args))
        return "(%s) |--> %s" % (args, repr_x)

    def _latex_element_(self, x):
        r"""
        Finds the LaTeX representation of this expression.

        EXAMPLES::

            sage: f(A, t, omega, psi) = A*cos(omega*t - psi)
            sage: f._latex_()
            '\\left( A, t, \\omega, \\psi \\right) \\ {\\mapsto} \\ A \\cos\\left(\\omega t - \\psi\\right)'

            sage: f(mu) =  mu^3
            sage: f._latex_()
            '\\mu \\ {\\mapsto}\\ \\mu^{3}'
        """
        from sage.misc.latex import latex
        args = self.args()
        args = [latex(arg) for arg in args]
        latex_x = SymbolicRing._latex_element_(self, x)
        if len(args) == 1:
            return r"%s \ {\mapsto}\ %s" % (args[0], latex_x)
        vars = ", ".join(args)
        return r"\left( %s \right) \ {\mapsto} \ %s" % (vars, latex_x)

    def _call_element_(self, _the_element, *args, **kwds):
        """
        Calling a callable symbolic expression returns a symbolic expression
        with the appropriate arguments substituted.

        EXAMPLES::

            sage: var('a, x, y, z')
            (a, x, y, z)
            sage: f(x,y) = a + 2*x + 3*y + z
            sage: f
            (x, y) |--> a + 2*x + 3*y + z
            sage: f(1,2)
            a + z + 8
            sage: f(y=2, a=-1)
            2*x + z + 5

        Note that keyword arguments will override the regular arguments.
        ::


            sage: f.arguments()
            (x, y)
            sage: f(1,2)
            a + z + 8
            sage: f(10,2)
            a + z + 26
            sage: f(10,2,x=1)
            a + z + 8
            sage: f(z=100)
            a + 2*x + 3*y + 100
        """
        if any(type(arg).__module__ == 'numpy' and type(arg).__name__ == "ndarray" for arg in args):  # avoid importing
            raise NotImplementedError("Numpy arrays are not supported as arguments for symbolic expressions")

        d = dict(zip([repr(_) for _ in self.arguments()], args))
        d.update(kwds)
        return SR(_the_element.substitute(**d))

    # __reduce__ gets replaced by the CallableSymbolicExpressionRingFactory
    __reduce__ = object.__reduce__


class CallableSymbolicExpressionRingFactory(UniqueFactory):
    def create_key(self, args, check=True):
        """
        EXAMPLES::

            sage: x,y = var('x,y')
            sage: CallableSymbolicExpressionRing.create_key((x,y))
            (x, y)
        """
        if check:
            from sage.structure.element import Expression
            if len(args) == 1 and isinstance(args[0], (list, tuple)):
                args, = args
            for arg in args:
                if not (isinstance(arg, Expression) and arg.is_symbol()):
                    raise TypeError("must construct a function with a tuple (or list) of variables")
            args = tuple(args)
        return args

    def create_object(self, version, key, **extra_args):
        """
        Return a CallableSymbolicExpressionRing given a version and a key.

        EXAMPLES::

            sage: x,y = var('x,y')
            sage: CallableSymbolicExpressionRing.create_object(0, (x, y))
            Callable function ring with arguments (x, y)
        """
        return CallableSymbolicExpressionRing_class(key)


CallableSymbolicExpressionRing = CallableSymbolicExpressionRingFactory('sage.symbolic.callable.CallableSymbolicExpressionRing')



# distutils: libraries = flint
# distutils: depends = flint/acb_mat.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void acb_mat_init(acb_mat_t mat, slong r, slong c) noexcept
    void acb_mat_clear(acb_mat_t mat) noexcept
    slong acb_mat_allocated_bytes(const acb_mat_t x) noexcept
    void acb_mat_window_init(acb_mat_t window, const acb_mat_t mat, slong r1, slong c1, slong r2, slong c2) noexcept
    void acb_mat_window_clear(acb_mat_t window) noexcept
    void acb_mat_set(acb_mat_t dest, const acb_mat_t src) noexcept
    void acb_mat_set_fmpz_mat(acb_mat_t dest, const fmpz_mat_t src) noexcept
    void acb_mat_set_round_fmpz_mat(acb_mat_t dest, const fmpz_mat_t src, slong prec) noexcept
    void acb_mat_set_fmpq_mat(acb_mat_t dest, const fmpq_mat_t src, slong prec) noexcept
    void acb_mat_set_arb_mat(acb_mat_t dest, const arb_mat_t src) noexcept
    void acb_mat_set_round_arb_mat(acb_mat_t dest, const arb_mat_t src, slong prec) noexcept
    void acb_mat_get_real(arb_mat_t re, const arb_mat_t mat) noexcept
    void acb_mat_get_imag(arb_mat_t im, const arb_mat_t mat) noexcept
    void acb_mat_set_real_imag(acb_mat_t mat, const arb_mat_t re, const arb_mat_t im) noexcept
    void acb_mat_randtest(acb_mat_t mat, flint_rand_t state, slong prec, slong mag_bits) noexcept
    void acb_mat_randtest_eig(acb_mat_t mat, flint_rand_t state, acb_srcptr E, slong prec) noexcept
    void acb_mat_printd(const acb_mat_t mat, slong digits) noexcept
    void acb_mat_fprintd(FILE * file, const acb_mat_t mat, slong digits) noexcept
    bint acb_mat_equal(const acb_mat_t mat1, const acb_mat_t mat2) noexcept
    bint acb_mat_overlaps(const acb_mat_t mat1, const acb_mat_t mat2) noexcept
    bint acb_mat_contains(const acb_mat_t mat1, const acb_mat_t mat2) noexcept
    bint acb_mat_contains_fmpz_mat(const acb_mat_t mat1, const fmpz_mat_t mat2) noexcept
    bint acb_mat_contains_fmpq_mat(const acb_mat_t mat1, const fmpq_mat_t mat2) noexcept
    bint acb_mat_eq(const acb_mat_t mat1, const acb_mat_t mat2) noexcept
    bint acb_mat_ne(const acb_mat_t mat1, const acb_mat_t mat2) noexcept
    bint acb_mat_is_real(const acb_mat_t mat) noexcept
    bint acb_mat_is_empty(const acb_mat_t mat) noexcept
    bint acb_mat_is_square(const acb_mat_t mat) noexcept
    bint acb_mat_is_exact(const acb_mat_t mat) noexcept
    bint acb_mat_is_zero(const acb_mat_t mat) noexcept
    bint acb_mat_is_finite(const acb_mat_t mat) noexcept
    bint acb_mat_is_triu(const acb_mat_t mat) noexcept
    bint acb_mat_is_tril(const acb_mat_t mat) noexcept
    bint acb_mat_is_diag(const acb_mat_t mat) noexcept
    void acb_mat_zero(acb_mat_t mat) noexcept
    void acb_mat_one(acb_mat_t mat) noexcept
    void acb_mat_ones(acb_mat_t mat) noexcept
    void acb_mat_onei(acb_mat_t mat) noexcept
    void acb_mat_indeterminate(acb_mat_t mat) noexcept
    void acb_mat_dft(acb_mat_t mat, int type, slong prec) noexcept
    void acb_mat_transpose(acb_mat_t dest, const acb_mat_t src) noexcept
    void acb_mat_conjugate_transpose(acb_mat_t dest, const acb_mat_t src) noexcept
    void acb_mat_conjugate(acb_mat_t dest, const acb_mat_t src) noexcept
    void acb_mat_bound_inf_norm(mag_t b, const acb_mat_t A) noexcept
    void acb_mat_frobenius_norm(arb_t res, const acb_mat_t A, slong prec) noexcept
    void acb_mat_bound_frobenius_norm(mag_t res, const acb_mat_t A) noexcept
    void acb_mat_neg(acb_mat_t dest, const acb_mat_t src) noexcept
    void acb_mat_add(acb_mat_t res, const acb_mat_t mat1, const acb_mat_t mat2, slong prec) noexcept
    void acb_mat_sub(acb_mat_t res, const acb_mat_t mat1, const acb_mat_t mat2, slong prec) noexcept
    void acb_mat_mul_classical(acb_mat_t res, const acb_mat_t mat1, const acb_mat_t mat2, slong prec) noexcept
    void acb_mat_mul_threaded(acb_mat_t res, const acb_mat_t mat1, const acb_mat_t mat2, slong prec) noexcept
    void acb_mat_mul_reorder(acb_mat_t res, const acb_mat_t mat1, const acb_mat_t mat2, slong prec) noexcept
    void acb_mat_mul(acb_mat_t res, const acb_mat_t mat1, const acb_mat_t mat2, slong prec) noexcept
    void acb_mat_mul_entrywise(acb_mat_t res, const acb_mat_t mat1, const acb_mat_t mat2, slong prec) noexcept
    void acb_mat_sqr_classical(acb_mat_t res, const acb_mat_t mat, slong prec) noexcept
    void acb_mat_sqr(acb_mat_t res, const acb_mat_t mat, slong prec) noexcept
    void acb_mat_pow_ui(acb_mat_t res, const acb_mat_t mat, ulong exp, slong prec) noexcept
    void acb_mat_approx_mul(acb_mat_t res, const acb_mat_t mat1, const acb_mat_t mat2, slong prec) noexcept
    void acb_mat_scalar_mul_2exp_si(acb_mat_t B, const acb_mat_t A, slong c) noexcept
    void acb_mat_scalar_addmul_si(acb_mat_t B, const acb_mat_t A, slong c, slong prec) noexcept
    void acb_mat_scalar_addmul_fmpz(acb_mat_t B, const acb_mat_t A, const fmpz_t c, slong prec) noexcept
    void acb_mat_scalar_addmul_arb(acb_mat_t B, const acb_mat_t A, const arb_t c, slong prec) noexcept
    void acb_mat_scalar_addmul_acb(acb_mat_t B, const acb_mat_t A, const acb_t c, slong prec) noexcept
    void acb_mat_scalar_mul_si(acb_mat_t B, const acb_mat_t A, slong c, slong prec) noexcept
    void acb_mat_scalar_mul_fmpz(acb_mat_t B, const acb_mat_t A, const fmpz_t c, slong prec) noexcept
    void acb_mat_scalar_mul_arb(acb_mat_t B, const acb_mat_t A, const arb_t c, slong prec) noexcept
    void acb_mat_scalar_mul_acb(acb_mat_t B, const acb_mat_t A, const acb_t c, slong prec) noexcept
    void acb_mat_scalar_div_si(acb_mat_t B, const acb_mat_t A, slong c, slong prec) noexcept
    void acb_mat_scalar_div_fmpz(acb_mat_t B, const acb_mat_t A, const fmpz_t c, slong prec) noexcept
    void acb_mat_scalar_div_arb(acb_mat_t B, const acb_mat_t A, const arb_t c, slong prec) noexcept
    void acb_mat_scalar_div_acb(acb_mat_t B, const acb_mat_t A, const acb_t c, slong prec) noexcept
    void _acb_mat_vector_mul_row(acb_ptr res, acb_srcptr v, const acb_mat_t A, slong prec) noexcept
    void _acb_mat_vector_mul_col(acb_ptr res, const acb_mat_t A, acb_srcptr v, slong prec) noexcept
    void acb_mat_vector_mul_row(acb_ptr res, acb_srcptr v, const acb_mat_t A, slong prec) noexcept
    void acb_mat_vector_mul_col(acb_ptr res, const acb_mat_t A, acb_srcptr v, slong prec) noexcept
    int acb_mat_lu_classical(slong * perm, acb_mat_t LU, const acb_mat_t A, slong prec) noexcept
    int acb_mat_lu_recursive(slong * perm, acb_mat_t LU, const acb_mat_t A, slong prec) noexcept
    int acb_mat_lu(slong * perm, acb_mat_t LU, const acb_mat_t A, slong prec) noexcept
    void acb_mat_solve_tril_classical(acb_mat_t X, const acb_mat_t L, const acb_mat_t B, int unit, slong prec) noexcept
    void acb_mat_solve_tril_recursive(acb_mat_t X, const acb_mat_t L, const acb_mat_t B, int unit, slong prec) noexcept
    void acb_mat_solve_tril(acb_mat_t X, const acb_mat_t L, const acb_mat_t B, int unit, slong prec) noexcept
    void acb_mat_solve_triu_classical(acb_mat_t X, const acb_mat_t U, const acb_mat_t B, int unit, slong prec) noexcept
    void acb_mat_solve_triu_recursive(acb_mat_t X, const acb_mat_t U, const acb_mat_t B, int unit, slong prec) noexcept
    void acb_mat_solve_triu(acb_mat_t X, const acb_mat_t U, const acb_mat_t B, int unit, slong prec) noexcept
    void acb_mat_solve_lu_precomp(acb_mat_t X, const slong * perm, const acb_mat_t LU, const acb_mat_t B, slong prec) noexcept
    int acb_mat_solve(acb_mat_t X, const acb_mat_t A, const acb_mat_t B, slong prec) noexcept
    int acb_mat_solve_lu(acb_mat_t X, const acb_mat_t A, const acb_mat_t B, slong prec) noexcept
    int acb_mat_solve_precond(acb_mat_t X, const acb_mat_t A, const acb_mat_t B, slong prec) noexcept
    int acb_mat_inv(acb_mat_t X, const acb_mat_t A, slong prec) noexcept
    void acb_mat_det_lu(acb_t det, const acb_mat_t A, slong prec) noexcept
    void acb_mat_det_precond(acb_t det, const acb_mat_t A, slong prec) noexcept
    void acb_mat_det(acb_t det, const acb_mat_t A, slong prec) noexcept
    void acb_mat_approx_solve_triu(acb_mat_t X, const acb_mat_t U, const acb_mat_t B, int unit, slong prec) noexcept
    void acb_mat_approx_solve_tril(acb_mat_t X, const acb_mat_t L, const acb_mat_t B, int unit, slong prec) noexcept
    int acb_mat_approx_lu(slong * P, acb_mat_t LU, const acb_mat_t A, slong prec) noexcept
    void acb_mat_approx_solve_lu_precomp(acb_mat_t X, const slong * perm, const acb_mat_t A, const acb_mat_t B, slong prec) noexcept
    int acb_mat_approx_solve(acb_mat_t X, const acb_mat_t A, const acb_mat_t B, slong prec) noexcept
    int acb_mat_approx_inv(acb_mat_t X, const acb_mat_t A, slong prec) noexcept
    void _acb_mat_charpoly(acb_ptr poly, const acb_mat_t mat, slong prec) noexcept
    void acb_mat_charpoly(acb_poly_t poly, const acb_mat_t mat, slong prec) noexcept
    void _acb_mat_companion(acb_mat_t mat, acb_srcptr poly, slong prec) noexcept
    void acb_mat_companion(acb_mat_t mat, const acb_poly_t poly, slong prec) noexcept
    void acb_mat_exp_taylor_sum(acb_mat_t S, const acb_mat_t A, slong N, slong prec) noexcept
    void acb_mat_exp(acb_mat_t B, const acb_mat_t A, slong prec) noexcept
    void acb_mat_trace(acb_t trace, const acb_mat_t mat, slong prec) noexcept
    void _acb_mat_diag_prod(acb_t res, const acb_mat_t mat, slong a, slong b, slong prec) noexcept
    void acb_mat_diag_prod(acb_t res, const acb_mat_t mat, slong prec) noexcept
    void acb_mat_get_mid(acb_mat_t B, const acb_mat_t A) noexcept
    void acb_mat_add_error_mag(acb_mat_t mat, const mag_t err) noexcept
    int acb_mat_approx_eig_qr(acb_ptr E, acb_mat_t L, acb_mat_t R, const acb_mat_t A, const mag_t tol, slong maxiter, slong prec) noexcept
    void acb_mat_eig_global_enclosure(mag_t eps, const acb_mat_t A, acb_srcptr E, const acb_mat_t R, slong prec) noexcept
    void acb_mat_eig_enclosure_rump(acb_t lmbda, acb_mat_t J, acb_mat_t R, const acb_mat_t A, const acb_t lambda_approx, const acb_mat_t R_approx, slong prec) noexcept
    int acb_mat_eig_simple_rump(acb_ptr E, acb_mat_t L, acb_mat_t R, const acb_mat_t A, acb_srcptr E_approx, const acb_mat_t R_approx, slong prec) noexcept
    int acb_mat_eig_simple_vdhoeven_mourrain(acb_ptr E, acb_mat_t L, acb_mat_t R, const acb_mat_t A, acb_srcptr E_approx, const acb_mat_t R_approx, slong prec) noexcept
    int acb_mat_eig_simple(acb_ptr E, acb_mat_t L, acb_mat_t R, const acb_mat_t A, acb_srcptr E_approx, const acb_mat_t R_approx, slong prec) noexcept
    int acb_mat_eig_multiple_rump(acb_ptr E, const acb_mat_t A, acb_srcptr E_approx, const acb_mat_t R_approx, slong prec) noexcept
    int acb_mat_eig_multiple(acb_ptr E, const acb_mat_t A, acb_srcptr E_approx, const acb_mat_t R_approx, slong prec) noexcept

from .acb_mat_macros cimport *



# sage.doctest: needs sage.libs.pari
r"""
List of coset representatives for `\Gamma_H(N)` in `\SL_2(\ZZ)`
"""
###########################################################################
#       Sage: Open Source Mathematical Software
#
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
###########################################################################
from sage.misc.persist import register_unpickle_override
from sage.modular.modsym import p1list
from sage.structure.richcmp import richcmp, richcmp_method
from sage.structure.sage_object import SageObject


@richcmp_method
class GHlist(SageObject):
    r"""
    A class representing a list of coset representatives for `\Gamma_H(N)` in
    `\SL_2(\ZZ)`.

    TESTS::

        sage: L = sage.modular.modsym.ghlist.GHlist(GammaH(18,[13]))
        sage: loads(dumps(L)) == L
        True
    """
    def __init__(self, group) -> None:
        """
        EXAMPLES::

            sage: L = sage.modular.modsym.ghlist.GHlist(GammaH(8,[7])); L # indirect doctest
            List of coset representatives for Congruence Subgroup Gamma_H(8) with H generated by [7]
        """
        self.__group = group
        N = group.level()
        v = group._coset_reduction_data()[0]
        N = group.level()
        coset_reps = {a for a, b, _ in v if b == 1}
        w = [group._reduce_coset(x * u, x * v)
             for x in coset_reps for u, v in p1list.P1List(N).list()]
        w = sorted(set(w))
        self.__list = w

    def __getitem__(self, i):
        """
        EXAMPLES::

            sage: L = sage.modular.modsym.ghlist.GHlist(GammaH(8, [5])); L[5] # indirect doctest
            (1, 3)
        """
        return self.__list[i]

    def __richcmp__(self, other, op):
        r"""
        Compare ``self`` to ``other``.

        EXAMPLES::

            sage: L1 = sage.modular.modsym.ghlist.GHlist(GammaH(18, [11]))
            sage: L2 = sage.modular.modsym.ghlist.GHlist(GammaH(18, [13]))
            sage: L1 > L2
            True
            sage: L1 == QQ
            False
        """
        if not isinstance(other, GHlist):
            return NotImplemented
        return richcmp(self.__group, other.__group, op)

    def __len__(self):
        """
        Return the length of the underlying list (the index of the group).

        EXAMPLES::

            sage: L = sage.modular.modsym.ghlist.GHlist(GammaH(24, [5])); len(L) # indirect doctest
            192
        """
        return len(self.__list)

    def __repr__(self):
        """
        String representation of ``self``.

        EXAMPLES::

            sage: L = sage.modular.modsym.ghlist.GHlist(GammaH(11, [3]))
            sage: L.__repr__()
            'List of coset representatives for Congruence Subgroup Gamma_H(11) with H generated by [3]'
        """
        return "List of coset representatives for %s" % self.__group

    def list(self):
        r"""
        Return a list of vectors representing the cosets. Do not change the
        returned list!

        EXAMPLES::

            sage: L = sage.modular.modsym.ghlist.GHlist(GammaH(4,[])); L.list()
            [(0, 1), (0, 3), (1, 0), (1, 1), (1, 2), (1, 3), (2, 1), (2, 3), (3, 0), (3, 1), (3, 2), (3, 3)]
        """
        return self.__list

    def normalize(self, u, v):
        r"""
        Given a pair `(u,v)` of integers, return the unique pair `(u', v')`
        such that the pair `(u', v')` appears in ``self.list()`` and `(u, v)`
        is equivalent to `(u', v')`.

        This will only make sense if `{\rm gcd}(u, v, N) = 1`; otherwise the
        output will not be an element of ``self``.

        EXAMPLES::

            sage: sage.modular.modsym.ghlist.GHlist(GammaH(24, [17, 19])).normalize(17, 6)
            (1, 6)
            sage: sage.modular.modsym.ghlist.GHlist(GammaH(24, [7, 13])).normalize(17, 6)
            (5, 6)
            sage: sage.modular.modsym.ghlist.GHlist(GammaH(24, [5, 23])).normalize(17, 6)
            (7, 18)
        """
        return self.__group._reduce_coset(u, v)


class _GHlist_old_pickle(GHlist):
    """
    This class exists only for dealing with old pickles.

    This needs to handle both old-style class pickles, where there is
    no input to the class on the initial ``__init__`` call, and the
    new class pickles, we need to have ``__setstate__`` handle it.
    """
    def __init__(self) -> None:
        """
        For unpickling old pickles.

        TESTS::

            sage: from sage.modular.modsym.ghlist import _GHlist_old_pickle
            sage: L = _GHlist_old_pickle()
            sage: type(L) == GHlist
            True
        """
        self.__class__ = GHlist

    def __setstate__(self, state):
        """
        For unpickling new pickles.

        TESTS::

            sage: from sage.modular.modsym.ghlist import GHlist
            sage: L = GHlist(GammaH(4,[]))
            sage: Lp = loads(dumps(L))
            sage: L == Lp
            True
            sage: type(Lp) == GHlist
            True
        """
        # We don't really want this class, but we want to handle new
        #   pickles without creating a new class
        self.__class__ = GHlist
        self.__dict__ = state  # Default pickling is ``state = self.__dict__``


register_unpickle_override('sage.modular.modsym.ghlist', 'GHlist',
                           _GHlist_old_pickle)



r"""
Class groups of binary quadratic forms

EXAMPLES:

Constructing the class of a given binary quadratic form is straightforward::

    sage: F1 = BinaryQF([22, 91, 99])
    sage: cl1 = F1.form_class(); cl1
    Class of 5*x^2 - 3*x*y + 22*y^2

Every class is represented by a reduced form in it::

    sage: cl1.form()
    5*x^2 - 3*x*y + 22*y^2
    sage: cl1.form() == F1.reduced_form()
    True

Addition of form classes and derived operations are defined by composition
of binary quadratic forms::

    sage: F2 = BinaryQF([4, 1, 27])
    sage: cl2 = F2.form_class(); cl2
    Class of 4*x^2 + x*y + 27*y^2
    sage: cl1 + cl2
    Class of 9*x^2 + x*y + 12*y^2
    sage: cl1 + cl2 == (F1 * F2).form_class()
    True
    sage: -cl1
    Class of 5*x^2 + 3*x*y + 22*y^2
    sage: cl1 - cl1
    Class of x^2 + x*y + 108*y^2

The form class group can be constructed as an explicit parent object::

    sage: F1.discriminant()
    -431
    sage: Cl = BQFClassGroup(-431); Cl
    Form Class Group of Discriminant -431
    sage: cl1.parent() is Cl
    True
    sage: Cl(F1) == cl1
    True

Structural properties of the form class group, such as the class number,
the group invariants, and element orders, can be computed::

    sage: Cl.order()
    21
    sage: cl1 * Cl.order() == Cl.zero()
    True
    sage: cl2 * Cl.order() == Cl.zero()
    True
    sage: cl2.order()
    7
    sage: cl2 * cl2.order() == Cl.zero()
    True
    sage: Cl.abelian_group()
    Additive abelian group isomorphic to Z/21 embedded in
    Form Class Group of Discriminant -431
    sage: Cl.gens()  # random
    [Class of 5*x^2 + 3*x*y + 22*y^2]
    sage: Cl.gens()[0].order()
    21

AUTHORS:

- Lorenz Panny (2023)
"""

# ****************************************************************************
#       Copyright (C) 2023 Lorenz Panny
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method

from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.element import AdditiveGroupElement
from sage.categories.morphism import Morphism

from sage.misc.prandom import randrange
from sage.rings.integer_ring import ZZ
from sage.rings.finite_rings.integer_mod import Mod
from sage.arith.misc import random_prime
from sage.groups.generic import order_from_multiple, multiple
from sage.groups.additive_abelian.additive_abelian_wrapper import AdditiveAbelianGroupWrapper
from sage.quadratic_forms.binary_qf import BinaryQF

from sage.libs.pari import pari


class BQFClassGroup(Parent, UniqueRepresentation):
    r"""
    This type represents the class group for a given discriminant `D`.

    - For `D < 0`, the group is the class group of *positive definite*
      binary quadratic forms. The "full" form class group is the direct
      sum of two isomorphic copies of this group (one for positive
      definite forms and one for negative definite forms).

    - For `D > 0`, this functionality is currently not implemented.

    EXAMPLES::

        sage: BQFClassGroup(-4)
        Form Class Group of Discriminant -4
        sage: BQFClassGroup(-6)
        Traceback (most recent call last):
        ...
        ValueError: not a discriminant

    The discriminant need not be fundamental::

        sage: BQFClassGroup(-22^2)
        Form Class Group of Discriminant -484
    """

    def __init__(self, D, *, check=True) -> None:
        r"""
        Construct the class group for a given discriminant `D`.

        TESTS:

        Check that positive discriminants are rejected until code is
        written for them::

            sage: BQFClassGroup(101)
            Traceback (most recent call last):
            ...
            NotImplementedError: positive discriminants are not yet supported
        """
        self._disc = ZZ(D)
        if check:
            if not self._disc or self._disc % 4 not in (0, 1):
                raise ValueError('not a discriminant')
            if self._disc > 0:
                raise NotImplementedError('positive discriminants are not yet supported')
        super().__init__()

    def _element_constructor_(self, F, *, check=True):
        r"""
        Construct an element of this form class group as a :class:`BQFClassGroup_element`.

        EXAMPLES::

            sage: Cl = BQFClassGroup(-999)
            sage: Cl(0)                         # indirect doctest
            Class of x^2 + x*y + 250*y^2
            sage: Cl(BinaryQF([16, 5, 16]))     # indirect doctest
            Class of 16*x^2 + 5*x*y + 16*y^2
        """
        if isinstance(F, BQFClassGroup_element):
            if check and F.parent() is not self:
                # class group is unique parent
                raise ValueError('quadratic form has incorrect discriminant')
            return F
        if F == 0:
            return self.zero()
        if check and not isinstance(F, BinaryQF):
            raise TypeError('not a binary quadratic form')
        return BQFClassGroup_element(F, parent=self, check=check)

    def zero(self):
        r"""
        Return the neutral element of this group.

        This is the class of the principal binary quadratic form of
        the respective discriminant.

        EXAMPLES::

            sage: Cl = BQFClassGroup(-999)
            sage: cl = Cl.zero(); cl
            Class of x^2 + x*y + 250*y^2
            sage: cl + cl == cl
            True
        """
        return self(BinaryQF.principal(self._disc))

    def random_element(self):
        r"""
        Return a somewhat random element of this form class group.

        ALGORITHM:

        Sample random odd primes `a` until `b^2 = D \pmod{4a}` has a
        solution `b \in \ZZ` and set `c = (b^2-D)/(4a)`. Flip a coin
        to choose the sign of `b`. Then return the class of `[a,b,c]`.

        .. NOTE::

            No strict guarantees are being made about the distribution of
            classes sampled by this function. Heuristically, however, it
            should be fairly close to uniform.

        EXAMPLES::

            sage: Cl = BQFClassGroup(-999); Cl
            Form Class Group of Discriminant -999
            sage: cl = Cl.random_element(); cl  # random
            Class of 10*x^2 + x*y + 25*y^2
            sage: cl.form().discriminant()
            -999
        """
        B = self._disc.abs() * 100 + 9999
        while True:
            a = random_prime(B, proof=False, lbound=3)
            if self._disc.kronecker(a) == 1:
                break
        b = ZZ(Mod(self._disc, 4*a).sqrt())
        c = (b**2 - self._disc) // (4*a)
        if randrange(2):
            b = -b
        return self(BinaryQF([a, b, c]))

    def __hash__(self) -> int:
        r"""
        Return a hash value for this form class group.

        EXAMPLES::

            sage: hash(BQFClassGroup(-999))  # random
            -4246560339810542104
        """
        return hash(('BQFClassGroup', self._disc))

    def _repr_(self) -> str:
        r"""
        Return a string describing this form class group.

        EXAMPLES::

            sage: BQFClassGroup(-999)  # indirect doctest
            Form Class Group of Discriminant -999
        """
        return f'Form Class Group of Discriminant {self._disc}'

    def discriminant(self):
        r"""
        Return the discriminant of the forms in this form class group.

        EXAMPLES::

            sage: BQFClassGroup(-999).discriminant()
            -999
        """
        return self._disc

    @cached_method
    def order(self):
        r"""
        Return the order of this form class group (the *class number*).

        ALGORITHM: :func:`sage.rings.number_field.order.quadratic_order_class_number`.

        EXAMPLES::

            sage: BQFClassGroup(-4).order()
            1
            sage: BQFClassGroup(-11).order()
            1
            sage: BQFClassGroup(-67).order()
            1
            sage: BQFClassGroup(-163).order()
            1
            sage: BQFClassGroup(-999).order()
            24
            sage: BQFClassGroup(-9999).order()
            88
            sage: BQFClassGroup(-99999).order()
            224
        """
        # Beware: If this code is ever generalized to positive
        # discriminants, care must be taken to use the correct
        # notion of class number. We may need the *narrow* class
        # number here; see PARI's documentation for qfbclassno().
        from sage.rings.number_field.order import quadratic_order_class_number
        return quadratic_order_class_number(self._disc)

    cardinality = order

    @cached_method
    def abelian_group(self):
        r"""
        Return the structure of this form class group as an
        :class:`AdditiveAbelianGroupWrapper` object.

        ALGORITHM: :pari:`quadclassunit`

        EXAMPLES::

            sage: Cl = BQFClassGroup(-4*777)
            sage: Cl.order()
            16
            sage: G = Cl.abelian_group(); G
            Additive abelian group isomorphic to Z/4 + Z/2 + Z/2 embedded in
            Form Class Group of Discriminant -3108
            sage: G.gens()  # random
            (Class of 11*x^2 + 4*x*y + 71*y^2,
             Class of 6*x^2 + 6*x*y + 131*y^2,
             Class of 2*x^2 + 2*x*y + 389*y^2)
            sage: [g.order() for g in G.gens()]
            [4, 2, 2]
            sage: G.discrete_log(Cl.random_element())  # random
            (3, 0, 1)
        """
        h, ords, gens, reg = pari.quadclassunit(self._disc)
        ords = [ZZ(o) for o in ords]
        gens = [BinaryQF(g) for g in gens]
        return AdditiveAbelianGroupWrapper(self, gens, ords)

    def gens(self) -> tuple:
        r"""
        Return a generating set of this form class group.

        EXAMPLES::

            sage: Cl = BQFClassGroup(-4*419)
            sage: Cl.gens()
            (Class of 3*x^2 + 2*x*y + 140*y^2,)

        ::

            sage: Cl = BQFClassGroup(-4*777)
            sage: Cl.gens()  # random
            (Class of 11*x^2 + 4*x*y + 71*y^2,
             Class of 6*x^2 + 6*x*y + 131*y^2,
             Class of 2*x^2 + 2*x*y + 389*y^2)
        """
        return tuple(g.element() for g in self.abelian_group().gens())

    def _coerce_map_from_(self, other):
        r"""
        Return the natural projection map between two class groups
        of binary quadratic forms when it is defined.

        .. SEEALSO:: :class:`BQFClassGroupQuotientMorphism`

        EXAMPLES::

            sage: G = BQFClassGroup(-4*117117)
            sage: H = BQFClassGroup(-4*77)
            sage: proj = G.hom(H); proj  # indirect doctest
            Coercion morphism:
              From: Form Class Group of Discriminant -468468
              To:   Form Class Group of Discriminant -308
            sage: elt = G(BinaryQF(333, 306, 422)); elt
            Class of 333*x^2 + 306*x*y + 422*y^2
            sage: proj(elt)
            Class of 9*x^2 + 4*x*y + 9*y^2
        """
        if not isinstance(other, BQFClassGroup):
            return super()._coerce_map_from_(other)
        try:
            proj = BQFClassGroupQuotientMorphism(other, self)
        except (TypeError, ValueError):
            return super()._coerce_map_from_(other)
        return proj


class BQFClassGroup_element(AdditiveGroupElement):
    r"""
    This type represents elements of class groups of binary quadratic forms.

    Users should not need to construct objects of this type directly; it can
    be accessed via either the :class:`BQFClassGroup` parent object or the
    :meth:`~BinaryQF.form_class` method associated to binary quadratic forms.

    Currently only classes of positive definite forms are supported.

    EXAMPLES::

        sage: F = BinaryQF([22, 91, 99])
        sage: F.form_class()  # indirect doctest
        Class of 5*x^2 - 3*x*y + 22*y^2

    ::

        sage: Cl = BQFClassGroup(-4*419)
        sage: Cl.zero()
        Class of x^2 + 419*y^2
        sage: Cl.gens()[0]  # indirect doctest
        Class of 3*x^2 + 2*x*y + 140*y^2
    """

    def __init__(self, F, parent, *, check=True, reduce=True) -> None:
        r"""
        Constructor for classes of binary quadratic forms.

        EXAMPLES::

            sage: Cl = BQFClassGroup(-431)
            sage: F = BinaryQF([22, 91, 99])
            sage: from sage.quadratic_forms.bqf_class_group import BQFClassGroup_element
            sage: BQFClassGroup_element(F, parent=Cl)
            Class of 5*x^2 - 3*x*y + 22*y^2
        """
        if check:
            if not isinstance(F, BinaryQF):
                raise TypeError('not a binary quadratic form')
            if F.discriminant() != parent.discriminant():
                raise ValueError('given quadratic form has wrong discriminant')
            if not F.is_primitive():
                raise ValueError('given quadratic form is not primitive')
            if not F.is_positive_definite():
                raise NotImplementedError('only positive definite forms are currently supported')
        if reduce:
            F = F.reduced_form()
        self._form = F
        super().__init__(parent=parent)

    def form(self):
        r"""
        Return a reduced quadratic form in this class.

        (For `D < 0`, each class contains a *unique* reduced form.)

        EXAMPLES::

            sage: F = BinaryQF([3221, 2114, 350])
            sage: cl = F.form_class()
            sage: cl.form()
            29*x^2 + 14*x*y + 350*y^2
            sage: cl.form() == F.reduced_form()
            True
        """
        return self._form

    def _neg_(self):
        r"""
        Return the inverse of this form class.

        The inverse class of a form `[a,b,c]` is represented by `[a,-b,c]`.

        EXAMPLES::

            sage: F = BinaryQF([11,21,31])
            sage: cl = F.form_class(); cl
            Class of 11*x^2 - x*y + 21*y^2
            sage: -cl                               # indirect doctest
            Class of 11*x^2 + x*y + 21*y^2
            sage: cl + (-cl) == cl.parent().zero()  # indirect doctest
            True
        """
        a, b, c = self._form
        F = BinaryQF([a, -b, c])
        return BQFClassGroup_element(F, parent=self.parent())

    def _add_(self, other):
        r"""
        Return the composition of two form classes.

        EXAMPLES::

            sage: cl1 = BinaryQF([11,21,31]).form_class(); cl1
            Class of 11*x^2 - x*y + 21*y^2
            sage: cl2 = BinaryQF([7,55,141]).form_class(); cl2
            Class of 7*x^2 - x*y + 33*y^2
            sage: cl1 + cl2                         # indirect doctest
            Class of 3*x^2 + x*y + 77*y^2
        """
        F = self._form * other._form
        return BQFClassGroup_element(F, parent=self.parent())

    def _sub_(self, other):
        r"""
        Return the composition of a form class with the inverse of another.

        EXAMPLES::

            sage: cl1 = BinaryQF([11,21,31]).form_class(); cl1
            Class of 11*x^2 - x*y + 21*y^2
            sage: cl2 = BinaryQF([7,55,141]).form_class(); cl2
            Class of 7*x^2 - x*y + 33*y^2
            sage: cl1 - cl2                         # indirect doctest
            Class of 9*x^2 - 7*x*y + 27*y^2
            sage: cl1 - cl2 == cl1 + (-cl2)         # indirect doctest
            True
        """
        return self + (-other)

    def __mul__(self, other):
        r"""
        Return an integer multiple of this form class with respect to
        repeated composition.

        ALGORITHM: :func:`multiple`

        EXAMPLES::

            sage: F = BinaryQF([11,21,31])
            sage: cl = F.form_class(); cl
            Class of 11*x^2 - x*y + 21*y^2
            sage: cl*0 == cl.parent().zero()        # indirect doctest
            True
            sage: cl*1 == cl                        # indirect doctest
            True
            sage: cl*(-1) == -cl                    # indirect doctest
            True
            sage: cl*2 == cl + cl                   # indirect doctest
            True
            sage: cl*5 == cl + cl + cl + cl + cl    # indirect doctest
            True
            sage: 5*cl == cl*5                      # indirect doctest
            True
            sage: cl*(-5) == -(5*cl)                # indirect doctest
            True
        """
        return multiple(self, other, operation='+')

    __rmul__ = __mul__

    def __eq__(self, other) -> bool:
        r"""
        Test two form classes for equality.

        EXAMPLES::

            sage: F = BinaryQF([11,21,31])
            sage: cl = F.form_class(); cl
            Class of 11*x^2 - x*y + 21*y^2
            sage: cl == cl      # indirect doctest
            True
            sage: -cl == cl     # indirect doctest
            False
        """
        # When generalizing to positive discriminants in the future, keep
        # in mind that for indefinite forms there can be multiple reduced
        # forms per class. This also affects the other comparison methods
        # as well as hashing.
        return self._form == other._form

    def __ne__(self, other) -> bool:
        r"""
        Test two form classes for inequality.

        EXAMPLES::

            sage: F = BinaryQF([11,21,31])
            sage: cl = F.form_class(); cl
            Class of 11*x^2 - x*y + 21*y^2
            sage: cl != cl      # indirect doctest
            False
            sage: -cl != cl     # indirect doctest
            True
        """
        return self._form != other._form

    def __lt__(self, other) -> bool:
        r"""
        Compare two form classes according to the lexicographic ordering
        on their coefficient lists.

        EXAMPLES::

            sage: cl1 = BinaryQF([7,55,141]).form_class(); cl1
            Class of 7*x^2 - x*y + 33*y^2
            sage: cl2 = BinaryQF([11,21,31]).form_class(); cl2
            Class of 11*x^2 - x*y + 21*y^2
            sage: cl1 < cl2     # indirect doctest
            True
            sage: cl1 > cl2     # indirect doctest
            False
        """
        return self._form < other._form

    def __bool__(self) -> bool:
        r"""
        Return ``True`` if this form class is *not* the principal class
        and ``False`` otherwise.

        EXAMPLES::

            sage: cl = BinaryQF([11,21,31]).form_class()
            sage: bool(cl)
            True
            sage: bool(0*cl)
            False
        """
        return self != self.parent().zero()

    def is_zero(self) -> bool:
        r"""
        Return ``True`` if this form class is the principal class and
        ``False`` otherwise.

        EXAMPLES::

            sage: cl = BinaryQF([11,21,31]).form_class()
            sage: cl.is_zero()
            False
            sage: (0*cl).is_zero()
            True
        """
        return not self

    def _repr_(self) -> str:
        r"""
        Return a string representation of this form class.

        EXAMPLES::

            sage: F = BinaryQF([11,21,31])
            sage: F.form_class()  # indirect doctest
            Class of 11*x^2 - x*y + 21*y^2
        """
        return f'Class of {self._form}'

    def __hash__(self) -> int:
        r"""
        Return a hash value for this form class.

        EXAMPLES::

            sage: cl = BinaryQF([11,21,31]).form_class()
            sage: hash(cl)  # random
            -7760578299759721732
        """
        return hash(('BQFClassGroup_element', self._form))

    @cached_method
    def order(self):
        r"""
        Return the order of this form class in its class group.

        ALGORITHM: :meth:`BQFClassGroup.order` and :func:`order_from_multiple`

        EXAMPLES::

            sage: cl = BinaryQF([11,21,31]).form_class()
            sage: cl.order()
            10
            sage: (cl+cl).order()
            5
            sage: (cl+cl+cl).order()
            10
            sage: (5*cl).order()
            2
        """
        return order_from_multiple(self, self.parent().cardinality())


class BQFClassGroupQuotientMorphism(Morphism):
    r"""
    Let `D` be a discriminant and `f > 0` an integer.

    Given the class groups `G` and `H` of discriminants `f^2 D` and `D`,
    this class represents the natural projection morphism `G \to H` which
    is defined by composing the class representative `[a,b,c]` with the
    principal form of the target discriminant.

    Alternatively, evaluating this map can be characterized as finding a
    class representative `[a,b,c]` satisfying `f^2 \mid a` and `f \mid b`
    and substituting `x \mapsto x/f`.

    This map is a well-defined group homomorphism.

    EXAMPLES::

        sage: from sage.quadratic_forms.bqf_class_group import BQFClassGroupQuotientMorphism
        sage: G = BQFClassGroup(-4*117117)
        sage: H = BQFClassGroup(-4*77)
        sage: proj = BQFClassGroupQuotientMorphism(G, H)
        sage: elt = G(BinaryQF(333, 306, 422))
        sage: proj(elt)
        Class of 9*x^2 + 4*x*y + 9*y^2

    TESTS:

    Check that it is really a group homomorphism::

        sage: D = -randrange(1, 10^4)
        sage: D *= 4 if D%4 not in (0,1) else 1
        sage: f = randrange(1, 10^3)
        sage: G = BQFClassGroup(f^2*D)
        sage: H = BQFClassGroup(D)
        sage: proj = G.hom(H)
        sage: proj(G.zero()) == H.zero()
        True
        sage: elt1 = G.random_element()
        sage: elt2 = G.random_element()
        sage: proj(elt1 + elt2) == proj(elt1) + proj(elt2)
        True

    Check that it satisfies compatibility::

        sage: ff = f * randrange(1, 10^3)
        sage: F = BQFClassGroup(ff^2*D)
        sage: proj = F.hom(H)
        sage: proj1 = F.hom(G)
        sage: proj2 = G.hom(H)
        sage: elt = F.random_element()
        sage: proj(elt) == proj2(proj1(elt))
        True
    """
    def __init__(self, G, H) -> None:
        r"""
        Initialize this morphism between class groups of binary
        quadratic forms.

        EXAMPLES::

            sage: from sage.quadratic_forms.bqf_class_group import BQFClassGroupQuotientMorphism
            sage: G = BQFClassGroup(-4*117117)
            sage: H = BQFClassGroup(-4*77)
            sage: f = BQFClassGroupQuotientMorphism(G, H)
            sage: TestSuite(f).run(skip='_test_category')
        """
        if not isinstance(G, BQFClassGroup):
            raise TypeError('G needs to be a BQFClassGroup')
        if not isinstance(H, BQFClassGroup):
            raise TypeError('H needs to be a BQFClassGroup')
        f2 = ZZ(G.discriminant() / H.discriminant())
        if not f2.is_square():
            raise ValueError('morphism only defined when disc(G) = f^2 * disc(H)')
        super().__init__(G, H)

    def _call_(self, elt):
        r"""
        Evaluate this morphism.

        EXAMPLES::

            sage: from sage.quadratic_forms.bqf_class_group import BQFClassGroupQuotientMorphism
            sage: G = BQFClassGroup(-4*117117)
            sage: H = BQFClassGroup(-4*77)
            sage: proj = BQFClassGroupQuotientMorphism(G, H)
            sage: elt = G(BinaryQF(333, 306, 422))
            sage: proj(elt)
            Class of 9*x^2 + 4*x*y + 9*y^2

        ALGORITHM: [Buell89]_, Theorem 7.9
        """
        one = BinaryQF.principal(self.codomain().discriminant())
        bqf = elt.form()
        bqf *= one
        return self.codomain()(bqf)
